from __future__ import annotations
from typing import Union, TypeVar, NamedTuple, Sequence, Deque, Optional, List, Type, Dict, Any
from types import ModuleType
from collections import namedtuple
import gc
import copy
import inspect
import sys
import xml.etree.ElementTree as ET
from xml.dom.minidom import Document
from xml.dom import minidom
import importlib
from collections import deque
import numpy as np
from numpy import ndarray

from .datatype import data_type
from .instanceattrs import InstanceAttrs
from .strcast import strcast
from .serializerutils import SerializerUtils
from .userinputnode import UserInputNode
from .serialize import Serialize
from .deserializeuserinput import DeserializeUserInput

#Generic class type alias
CLASS = TypeVar('CLASS')

#Generic class inst type alias
INSTANCE = TypeVar('INSTANCE')

class Deserialize(DeserializeUserInput):

    def set_path(
            self, 
            path: List, 
            index: int,
        ) -> List:
        child_path = copy.deepcopy(path)
        child_path.append(index)
        return child_path

    def tag_namedtuple(
            self, 
            node,
            ):
        node.tag = (
            node.tag 
            + '___' 
            + self.get_xml_node_attr(node, 'namedtuple_name')
        )
        self.set_xml_node_attr(node, 'type', 'dict')
        return node

    def tag_child_namedtuples(
            self, 
            node,
            ):
        for child in node:
            if self.is_namedtuple(child):
                self.tag_namedtuple(child)

    def is_namedtuple(
            self, 
            node,
            ):
        if self.attr_type(node) == 'namedtuple':
            return True

    def deserialize_nested_attr(
            self, 
            name, 
            val, 
            path,
            ):
        if isinstance(path, list):
            path = deque(path)
        return self.deserialize_nested_attr_(
            self.get_instance_dict(self)[name], 
            val, 
            path
        )

    def deserialize_nested_attr_(
            self, 
            attr, 
            val, 
            path,
            ):
        if len(path) > 1:
            pos = path.popleft()
            return self.deserialize_nested_attr_(
                attr[pos], 
                val, 
                path
            )
        attr[path[0]] = val
        return attr[path[0]] 
       
    def deserialize_list(
            self, 
            node, 
            child = None, 
            path = [],
            ):
        if len(path) != 0:
            null_list = [None for gchild in child]
            name = node.tag
            self.deserialize_nested_attr(name, null_list, path)
            for index, grandchild in enumerate(child):
                grandchild_path = self.set_path(path, index) 
                self.deserialize_child(node, grandchild, grandchild_path)
        else:
            null_list = self.null_adt(size=len(node)) 
            name = node.tag
            setattr(self, name, null_list)
            for index, child in enumerate(node):
                child_path = self.set_path(path, index) 
                self.deserialize_child(node, child, child_path)

    def list_to_tuple(
            self,
            val: Any
            ) -> Tuple:
        return (
            tuple(map(self.list_to_tuple, val)) 
                if type(val) == tuple or type(val) == list 
                else val 
        )

    def list_to_ndarray(
            self,
            val: Any
            ) -> Tuple:
        return (
            np.array(list(map(self.list_to_tuple, val))) 
                if type(val) == tuple or type(val) == list 
                else val 
        )
    
    def immutable_is_deserialized(
            self,
            tmp_tuple_attr_list: Union[List, np.ndarray],
            ) -> bool:
        if isinstance(tmp_tuple_attr_list, ndarray):
            tmp_tuple_attr_list = tmp_tuple_attr_list.tolist()
        if None in list(self.flatten_container(
            tmp_tuple_attr_list
        )):
            return False
        return True

    def flatten_container(
        self,
        container: Union[List, Tuple, Set],
        ) -> Generator[List, Tuple, Set]:
        for element in container:
            if isinstance(element, (list,tuple)):
                for nested_element in self.flatten_container(
                    element
                ):
                    yield nested_element
            else:
                yield element

    def deserialize_tuple(
            self, 
            node, 
            child = None, 
            path = [],
            ):
        if len(path) != 0:
            null_list = [None for gchild in child]
            name = node.tag
            self.deserialize_nested_attr(name, null_list, path)
            for index, grandchild in enumerate(child):
                grandchild_path = self.set_path(path, index) 
                self.deserialize_child(node, grandchild, grandchild_path)
            tmp_tuple_attr_list = getattr(self, name)
        else:
            null_list = self.null_adt(size=len(node)) 
            name = node.tag
            setattr(self, name, null_list)
            for index, child in enumerate(node):
                child_path = self.set_path(path, index) 
                self.deserialize_child(node, child, child_path)
        tmp_tuple_attr_list = getattr(self, name)
        if self.immutable_is_deserialized(tmp_tuple_attr_list):
            setattr(self, name, self.list_to_tuple(tmp_tuple_attr_list))

    def deserialize_ndarray(
            self, 
            node, 
            child = None, 
            path = [],
            ):
        if len(path) != 0:
            null_list = [None for gchild in child]
            name = node.tag
            self.deserialize_nested_attr(name, null_list, path)
            for index, grandchild in enumerate(child):
                grandchild_path = self.set_path(path, index) 
                self.deserialize_child(node, grandchild, grandchild_path)
            tmp_ndarray_attr_list = getattr(self, name)
        else:
            null_list = self.null_adt(size=len(node)) 
            name = node.tag
            setattr(self, name, null_list)
            for index, child in enumerate(node):
                child_path = self.set_path(path, index) 
                self.deserialize_child(node, child, child_path)
        tmp_ndarray_attr_list = getattr(self, name)
        if self.immutable_is_deserialized(tmp_ndarray_attr_list):
            setattr(self, name, self.list_to_ndarray(tmp_ndarray_attr_list))
            
            

    def deserialize_dict(
            self, 
            node, 
            child = None, 
            path = [],
            ):
        if len(path) != 0:
            name = node.tag
            null_dict = self.null_adt(
                keys=child, 
                key_attr='tag'
            )  
            self.deserialize_nested_attr(name, null_dict, path)
            for grandchild in child:
                grandchild_path = self.set_path(path, grandchild.tag) 
                self.deserialize_child(node, grandchild, grandchild_path)
        else:
            null_dict = self.null_adt(
                keys=node, 
                key_attr='tag'
            )  
            name = node.tag
            setattr(self, name, null_dict)
            for child in node:
                child_path = self.set_path(path, child.tag) 
                self.deserialize_child(node, child, child_path)

    def deserialize_namedtuple(
            self, 
            node, 
            child = None, 
            path = [],
            ):
        self.namedtuple_to_dict(node, child, path)
        self.dict_to_namedtuple(self)

    def is_namedtuple_dict(
            self, 
            name,
            ):
        if '___' in name:
            return True

    def namedtuple_name(
            self, 
            dict_name,
            ):
        name = dict_name.split('___')[0]
        return name

    def namedtuple_attrs(
            self, 
            cls,
            ):
        namedtuple_attr_names = []
        non_namedtuple_attr_names = []
        for attr_name in self.get_instance_dict(cls):
            if self.is_namedtuple_dict(attr_name):
                namedtuple_attr_names.append(attr_name)
            else:
                non_namedtuple_attr_names.append(attr_name)
        return namedtuple_attr_names, non_namedtuple_attr_names
            

    def namedtuple_to_dict(
            self, 
            node, 
            child = None, 
            path = [],
            ):
        name = node.tag
        if self.is_namedtuple(node):
            name = self.tag_namedtuple(node) 
        if len(path) != 0:
            self.tag_child_namedtuples(child)
            null_dict = self.null_adt(
                keys=child, 
                key_attr='tag'
            )  
            self.deserialize_nested_attr(name, null_dict, path)
            for grandchild in child:
                grandchild_path = self.set_path(path, grandchild.tag) 
                self.deserialize_child(node, grandchild, grandchild_path)
        else:
            name = node.tag
            self.tag_child_namedtuples(node)
            null_dict = self.null_adt(
                keys=node, 
                key_attr='tag'
            )  
            setattr(self, name, null_dict)
            for child in node:
                child_path = self.set_path(path, child.tag) 
                self.deserialize_child(node, child, child_path)

    def dict_to_namedtuple(
            self, 
            mapping, 
            name = '',
            ):
        if data_type(mapping) == 'instance':
            namedtuple_attrs, not_namedtuple_attrs \
                = self.namedtuple_attrs(mapping)  
            for key in namedtuple_attrs:
                setattr(
                    mapping, 
                    key.split('___')[0], 
                    self.dict_to_namedtuple(
                        getattr(mapping, key), 
                        key
                    )
                )
        if isinstance(mapping, dict):
            for key in list(mapping):
                if '___' in key:
                    mapping[key.split('___')[0]] \
                        = self.dict_to_namedtuple(mapping[key], key)
                else:
                    mapping[key] \
                        = self.dict_to_namedtuple(mapping[key], key)
            return self.namedtuple_from_mapping(mapping, name)
        return mapping

    def namedtuple_from_mapping(
            self, 
            mapping, 
            name,
            ):
        if self.is_namedtuple_dict(name):
            mapping_keys = [key.split('___')[0] for key in list(mapping)]
            mapping_values = []
            for key in mapping_keys:
                mapping_values.append(mapping[key])
            mapping_ = dict(zip(mapping_keys, mapping_values))
            this_namedtuple_maker = namedtuple(name.split('___')[1], mapping_)
            return this_namedtuple_maker(**mapping_)
        else:
            return mapping
         

    def deserialize_instance(
            self, 
            node, 
            child = None, 
            path = [],
            ):
        if len(path) != 0:
            name = node.tag 
            class_inst = self.class_instance(child)
            self.deserialize_nested_attr(name, class_inst, path)
            class_inst._deserialize(child)
        else:
            class_inst = self.class_instance(node)
            setattr(self, node.tag, class_inst)
            class_inst._deserialize(node)
        
    def attr_type(
            self, 
            node,
            ):
        if len(getattr(node, 'attrib')) == 0:
            return 'NoneType'
        return node.attrib['type']

    def deserialize_child(
            self, 
            node, 
            child = None, 
            path = [],
            ):
        deserialize_attr_dict = {
            'int' : self.deserialize_int,
            'float' : self.deserialize_float,
            'float64' : self.deserialize_float64,
            'float32' : self.deserialize_float32,
            'int64' : self.deserialize_int64,
            'int32' : self.deserialize_int32,
            'complex64' : self.deserialize_complex64,
            'complex128' : self.deserialize_complex128,
            'str' : self.deserialize_str,
            'bool' : self.deserialize_bool,
            'instance' : self.deserialize_instance,
            #Added for legacy support
            'class' : self.deserialize_instance, 
            'list' : self.deserialize_list,
            'tuple' : self.deserialize_tuple,
            'ndarray' : self.deserialize_ndarray,
            'dict' : self.deserialize_dict,
            'namedtuple' : self.deserialize_namedtuple,
            'NoneType' : self.deserialize_nonetype
        }
        if len(path) != 0:
            typ = self.attr_type(child)
            return deserialize_attr_dict[typ](node, child, path)
        typ = self.attr_type(node)
        return deserialize_attr_dict[typ](node, child, path)

    def deserialize_int(
            self, 
            node, 
            child = None, 
            path = [],
            ):
        if path:
            name = node.tag
            val = int(child.text)
            return self.deserialize_nested_attr(name, val, path)
        name = node.tag
        val = int(node.text)
        setattr(self, name, val)

    def deserialize_nonetype(
            self, 
            node, 
            child = None, 
            path = [],
            ):
        if path:
            name = node.tag
            val = None 
            return self.deserialize_nested_attr(name , val, path)
        name = node.tag
        val = None 
        setattr(self, name, val)

    def deserialize_float(
            self, 
            node, 
            child = None, 
            path = [],
            ):
        if path:
            name = node.tag
            val = float(child.text)
            return self.deserialize_nested_attr(name , val, path)
        name = node.tag
        val = float(node.text)
        setattr(self, name, val)

    def deserialize_float64(
            self, 
            node, 
            child = None, 
            path = [],
            ):
        if path:
            name = node.tag
            val = np.float64(child.text)
            return self.deserialize_nested_attr(name , val, path)
        name = node.tag
        val = np.float64(node.text)
        setattr(self, name, val)

    def deserialize_complex64(
            self, 
            node, 
            child = None, 
            path = [],
            ):
        if path:
            name = node.tag
            val = np.complex64(child.text)
            return self.deserialize_nested_attr(name , val, path)
        name = node.tag
        val = np.complex64(node.text)
        setattr(self, name, val)

    def deserialize_complex128(
            self, 
            node, 
            child = None, 
            path = [],
            ):
        if path:
            name = node.tag
            val = np.complex128(child.text)
            return self.deserialize_nested_attr(name , val, path)
        name = node.tag
        val = np.complex128(node.text)
        setattr(self, name, val)

    def deserialize_float32(
            self, 
            node, 
            child = None, 
            path = [],
            ):
        if path:
            name = node.tag
            val = np.float32(child.text)
            return self.deserialize_nested_attr(name , val, path)
        name = node.tag
        val = np.float32(node.text)
        setattr(self, name, val)

    def deserialize_int64(
            self, 
            node, 
            child = None, 
            path = [],
            ):
        if path:
            name = node.tag
            val = np.int64(child.text)
            return self.deserialize_nested_attr(name , val, path)
        name = node.tag
        val = np.int64(node.text)
        setattr(self, name, val)

    def deserialize_int32(
            self, 
            node, 
            child = None, 
            path = [],
            ):
        if path:
            name = node.tag
            val = np.int32(child.text)
            return self.deserialize_nested_attr(name , val, path)
        name = node.tag
        val = np.int32(node.text)
        setattr(self, name, val)

    def deserialize_bool(
            self, 
            node, 
            child = None, 
            path = [],
            ):
        if path:
            name = node.tag
            val = strcast(child.text)
            return self.deserialize_nested_attr(name, val, path)
        name = node.tag
        val = strcast(node.text)
        setattr(self, name, val)

    def set_str(
            self, 
            s,
            ):
        if s == 'None':
            return ''
        return s
        
    def deserialize_str(
            self, 
            node, 
            child = None, 
            path = [],
            ):
        if path:
            name = node.tag
            val = self.set_str(str(child.text))
            return self.deserialize_nested_attr(name, val, path)
        name = node.tag
        val = self.set_str(str(node.text))
        setattr(self, name, val)

    def _deserialize(
            self, 
            root, 
            path = [],
            ):
        for child in root:
            if 'type' in child.attrib:
                self.deserialize_child(child)
            else:
                setattr(self, child.tag, None)
                            
    def deserialize(
            self, 
            root, 
            user_input = False,
            ):
        xml_root = ET.parse(root).getroot()
        if user_input:
            inst_attrs = InstanceAttrs(self)
            empty_str_attrs = []
            for attr in inst_attrs.default_instance_attrs():
                if inst_attrs.is_empty_str(attr):
                    empty_str_attrs.append(attr)
            self.deserialize_user_input(
                root, 
                xml_root, 
                empty_str_attrs, 
                inst_attrs
            )
            return
        self._deserialize(xml_root)
                    

'''
-----------------------------------------------------------------------------------------------------------------------------------------------
-----------------------------------------------------------------------------------------------------------------------------------------------
'''

class Serializer(Deserialize, Serialize):
    pass




import numpy as np
from collections import deque
import re
import sys
from typing import List, Tuple, Optional, Union, Dict, Type


class ListStrCastNode(object): 

    def __init__(
            self,
            start,
            end,
            parent = None,
            path = [],
            ):
        self.start = start
        self.end = end
        self.parent = parent
        self.utils = StrCastUtils()
        self.path = path 

    def top_child(
            self,
            children,
            ):
        return list(children)[-1] 

    def pop_child(
            self,
            children,
            pos,
            ):
        return children.pop(pos) 

    def is_child(
            self,
            start1,
            end1,
            start2,
            end2,
            ):
        if start1 > start2 and end1 < end2:
            return True
        return

    def is_leaf(
            self,
            children,
            pos,
            ):
        if self.top_child(children) == pos:
            return True
        return

    def is_chars_leaf(
            self,
            chars_dict,
            pos,
            ):
        if self.top_child(chars_dict) == pos:
            return True
        return

    def get_children(
            self,
            internal_nodes = True,
            leaves = True,
            ):
        if internal_nodes and leaves:
            return [
                child for name, child in self.__dict__.items() 
                    if 'child' in name or 'leaf' in name
            ]
        elif internal_nodes and not leaves:
            return [
                child for name, child in self.__dict__.items() 
                    if 'child' in name
            ]
        elif leaves and not internal_nodes:
            return [
                child for name, child in self.__dict__.items() 
                    if 'leaf' in name
            ]
        return

    def child_names(
            self, 
            node,
            ):
        return [
            name for name in self.__dict__ 
                if 'child' in name
        ]

    def leaf_names(
            self,
            node,
            ):
        return [
            name for name in self.__dict__
                if 'leaf' in name
        ]

    def child_path(self):
        path = [] 
        if self.path:
            path = [pos for pos in self.path] 
        path.append(len(self.get_children(self)))
        return path

    def add_child(
            self, 
            start,
            end,
            ):
        if self.child_names(self):
            name = (
                'child'
                + str(int(re.findall(
                    r'\d+', 
                    self.child_names(self)[-1])[0]) 
                + 1)
            )
        else:
            name = 'child0'
        setattr(
            self, name, ListStrCastNode(
                start, end, self,
                self.child_path()
            )
        ) 
        return getattr(self, name) 

    def add_leaf(
            self,
            start,
            end, 
            s,
            chars = False
            ):
        name = None
        if self.leaf_names(self):
            name = (
                'leaf'
                + str(int(re.findall(
                    r'\d+', 
                    self.leaf_names(self)[-1])[0])
                + 1)
            )
        else:
            name = 'leaf0'
        leaf = self.utils.cleanup_str(
            s, start + 1,
            end, 'list'
        )
        if chars:
            setattr(self, name, tuple(leaf))
        else:
            setattr(self, name, leaf)
        return getattr(self, name) 
    def is_numpy_dtype(
            self,
            val: str,
            ) -> bool:
        numpy_dtype_suffixes = (
            '_INT32',
            '_INT64',
            '_FLOAT32',
            '_FLOAT64',
            '_COMPLEX64',
            '_COMPLEX128',
        )
        for numpy_dtype_suffix in numpy_dtype_suffixes:
            if numpy_dtype_suffix in val:
                return True
        return False

class StrCastUtils(object):

    def is_numpy_dtype(
            self,
            val: str,
            ) -> bool:
        numpy_dtype_suffixes = (
            '_INT32',
            '_INT64',
            '_FLOAT32',
            '_FLOAT64',
            '_COMPLEX64',
            '_COMPLEX128',
        )
        for numpy_dtype_suffix in numpy_dtype_suffixes:
            if numpy_dtype_suffix in val:
                return True
        return False

    def parse_numpy_dtype(
            self,
            val: str,
            ) -> Tuple[str]:
        dtype_name = \
            val.partition('_')[-1].replace('_', '').lower() 
        parsed_val = val.partition('_')[0]
        return parsed_val, dtype_name

    def xml_str_cast(
            self,
            val,
            ):
        if type(val.text) == type(None):
            return   
        else:
            if val.text.lower() == 'true':
                return True 
            elif val.text.lower() == 'false':
                return False 
            elif '[' in val.text:
                cast = StrCast()
                return cast.str_to_list(val)
            elif self.is_numpy_dtype(val.text):
                parsed_str_val, dtype_name = \
                    self.parse_numpy_dtype(val.text)
                return self.__str_cast(val.text, dtype_name)             
            elif val.text == '.' or re.search('[a-zA-Z]', val.text):
                return self.__str_cast(val.text, 'str')
            elif re.search('[\.]', val.text):
                return self.__str_cast(val.text, 'float')
            else:
                return self.__str_cast(val.text, 'int')

    def std_str_cast(
            self,
            val,
            ):
        if type(val) == type(None):
            return   
        else:
            if val.lower() == 'true':
                return True 
            elif val.lower() == 'false':
                return False 
            elif '[' in val:
                cast = StrCast()
                return cast.str_to_list(val)
            elif self.is_numpy_dtype(val):
                parsed_val, dtype_name = \
                    self.parse_numpy_dtype(val)
                return self.__str_cast(parsed_val, dtype_name)             
            elif val == '.' or re.search('[a-zA-Z]', val):
                return self.__str_cast(val, 'str')
            elif re.search('[\.]', val):
                return self.__str_cast(val, 'float')
            else:
                return self.__str_cast(val, 'int')

    def str_cast(
            self, 
            val,
            ):
        if hasattr(val, 'text'):
            return self.xml_str_cast(val)
        else:
            return self.std_str_cast(val)

    def __str_cast(
            self, 
            str_val,
            type_str,
            ):
        type_lookup = {
            'int' : int,
            'float' : float,
            'str' : str,
            'int32' : np.int32,
            'int64' : np.int64,
            'float32' : np.float32,
            'float64' : np.float64,
            'complex64' : np.complex64,
            'complex128' : np.complex128
        } 
        return type_lookup[type_str](str_val)

    def except_the_kitchen_sink_regex(
            self, 
            str_struct
            ):
        struct = re.findall(
            r'\"[^\"]*\"|\S*[A-Za-z]\S*[0-9]\S*|'
            '[0-9]?\S*[A-Za-z]\S*|'
            '-?\.?\d+\.?\d*|\d+',
            str_struct
        )
        return struct

    def str_cast_list_elements(
            self,
            str_list,
            ):
        parsed_str_list = self.except_the_kitchen_sink_regex(
            str_list
        )
        return self.cleanup_str_list(
            parsed_str_list 
        )

    def cleanup_str_list(
            self,
            parsed_str_list,
            ) -> List[str]:
        parsed_str_list = [
            list_element if '"' in list_element 
            else list_element.replace(',', '') 
                for list_element in parsed_str_list
        ]
        parsed_str_list = [
            list_element.replace('"', '') for list_element
            in parsed_str_list
        ]
        parsed_str_list = [
            self.str_cast(list_element) for list_element
            in parsed_str_list
        ]
        return parsed_str_list
        

    def str_cast_dict_elements(
            self,
            str_dict,
            ): #Come back to this later
        return  

#get rid of *args
    def str_cast_struct_elements(
            self,
            str_struct,
            *args,
            ):
        str_cast_struct_dict = {
            'list' : self.str_cast_list_elements(str_struct), 
            'dict' : self.str_cast_dict_elements(str_struct),
        }
        struct_type = args[0]
        return str_cast_struct_dict[struct_type]

#get rid of *args
    """
    def find_matching(
            self,
            s,
            *args,
            ):
        matching_dict = {}
        stack = []
        start, end = args
        for pos, i in enumerate(s):
            if i == start:
                stack.append(pos)
            elif i == end:
                matching_dict[stack.pop()] = pos
        return matching_dict
    """

    def is_start_end(self, char, start, end, matching_dict):
        if start == end:
            if char == start:
                if None in matching_dict.values():
                    return 'end'
                return 'start'
            return ''
        if char == start:
            return 'start'
        elif char == end:
            return 'end'
        return ''

    def find_matching(
            self, 
            s,
            start,
            end,
            ):
        matching_dict = {}
        stack = []
        for pos, i in enumerate(s):
            if self.is_start_end(i, start, end, matching_dict) == "start":
                matching_dict[pos] = None
                stack.append(pos)
            elif self.is_start_end(i, start, end, matching_dict) == "end":
                matching_dict[stack.pop()] = pos
        return matching_dict

    def num_lists(
            self,
            s,
            ):
        return self.num_structs(s, 'list')

    def num_dicts(
            self,
            s,
            ):
        return self.num_structs(s, 'dict')

    def num_structs(
            self,
            s,
            struct_type,
            ):
        num_structs_dict = { 
            'list' : self.find_matching(s, '[', ']'), 
            'dict' : self.find_matching(s, '{', '}')
        }
        return len(num_structs_dict[struct_type])

    def nested_list_insert(
            self,
            lst,
            val,
            path,
            ):
        if isinstance(path, list):
            path = deque(path)
        self.__nested_list_insert(lst, val, path)
        return lst

    def __nested_list_insert(
            self,
            lst,
            val,
            path,
            ):
        if path:
            pos = path.popleft()
            return self.__nested_list_insert(lst[pos], val, path)
        lst.append(val)

    def sort_dict(
            self,
            d,
            *args,
            **kwargs,
            ):
        args = list(args)
        for index, arg in enumerate(args):
            if arg == 'key':
                args[index] = 0
            if arg == 'value':
                args[index] = 1
        kwargs_keys = [key for key, value in kwargs.items()]
        return {
            key : value for key, value in sorted(
                d.items(), key = lambda item : item[args[0]], 
                reverse = kwargs[kwargs_keys[0]]
            )
        }

    def cleanup_str(
            self, 
            s, 
            start, 
            end, 
            *args,
            ):
        return self.str_cast_struct_elements(s[start : end], *args)

    def chars(
            self, 
            s, 
            *args,
            ):
        chars_dict = {}
        prev_bracket = {}
        for index, char in enumerate(s):
            if char == args[0]:
                if prev_bracket:
                    start, end = list(prev_bracket)[-1], index
                    if prev_bracket[start] == args[0]:
                        if (end - start) != 1:
                            chars_dict[start] = self.cleanup_str(
                                s, start, end,
                                args[2]
                            )
                    elif prev_bracket[start] == args[1]:
                        if (end - start) != 3:
                            chars_dict[start] = self.cleanup_str(
                                    s, start, end,
                                    args[2]
                                )
                prev_bracket = {index : args[0]} 
            elif char == args[1]:
                start, end = list(prev_bracket)[-1], index
                if prev_bracket[start] == args[1]:
                    if (end - start) != 1:
                        chars_dict[start] = self.cleanup_str(
                            s, start, end,
                            args[2]
                        )
                prev_bracket = {index : args[1]} 
        return chars_dict

    
class StrCast(object):

    def __init__(self):
        self.utils = StrCastUtils()

    def adjust_recursion_limit(
            self,
            s,
            ):
        if self.utils.num_lists(s) > 100:
            sys.setrecursionlimit(int(10e8))

    def start_end(
            self, 
            q_s, 
            pos = None,
            ):
        if not pos:
            start = list(q_s[-1])[0]
            end = q_s[-1][start]
            return start, end 
        start = list(q_s[pos])[0]
        end = q_s[pos][start]
        return start, end 
        
    def list_init_queue(
            self, 
            s,
            ):
        return deque([
            {start : end} for start, end in self.utils.sort_dict(
                self.utils.find_matching(s, '[', ']'),
                'key', reverse = True).items()
        ])

    def list_init_children(
            self, 
            s,
            ):
        return self.utils.sort_dict(
            self.utils.find_matching(s, '[', ']'),
            'value', reverse = True
        )

    def list_init_stack(
            self,
            s,
            init_queue = None,
            ):
        if init_queue:
            return deque([init_queue.pop()])
        else:
            return deque()

    def list_str_tree(
            self,
            s,
            **qscpn,
            ):
        if not qscpn:
            self.adjust_recursion_limit(s)
            qscpn['queue'] = self.list_init_queue(s)
            if len(qscpn['queue']) == 1:
                qs, qe = self.start_end(qscpn['queue']) 
                node = ListStrCastNode(qs, qe)
                node.add_leaf(qs, qe, s)
                return node
            qscpn['stack'] = self.list_init_stack(s, qscpn['queue'])
            qscpn['children'] = self.list_init_children(s) 
            qscpn['pos'] =  self.start_end(qscpn['queue'])[0]
            qscpn['node'] = ListStrCastNode(*self.start_end(qscpn['stack']))
            qscpn['node'].pop_child(qscpn['children'], 0)
        queue, stack, children, pos, node = qscpn.values()
        qs, qe = self.start_end(queue) 
        ss, se = self.start_end(stack) 
        if len(queue) > 1:
            if se > qe:
                if node.is_leaf(children, pos):
                    node.pop_child(children, qs)
                    node.add_leaf(qs, qe, s)
                    queue.pop()
                    pos = self.start_end(queue)[0]
                    qscpn = dict(zip(
                        qscpn.keys(), (
                            queue,
                            stack, 
                            children, 
                            pos, 
                            node,
                        )
                    ))
                    self.list_str_tree(s, **qscpn)
                else:
                    node.pop_child(children, qs)
                    child = node.add_child(qs, qe)
                    stack.append(queue.pop())
                    pos = self.start_end(queue)[0]
                    qscpn = dict(zip(
                        qscpn.keys(), (
                            queue, 
                            stack, 
                            children,
                            pos,
                            child,
                        )
                    ))
                    self.list_str_tree(s, **qscpn)
            elif qe > se:
                stack.pop()
                qscpn = dict(zip(
                    qscpn.keys(), (
                        queue, 
                        stack, 
                        children, 
                        pos, 
                        node.parent,
                    )
                ))
                self.list_str_tree(s, **qscpn)
        if children:
            if node.is_child(
                    qs, qe, ss, 
                    se
                ):
                if node.is_leaf(children, pos):
                    node.pop_child(children, qs)
                    node.add_leaf(qs, qe, s)
                    queue.pop()
            else:
                stack.pop()
                qscpn = dict(zip(
                    qscpn.keys(), ( 
                        queue, 
                        stack,
                        children, 
                        pos, 
                        node.parent
                    )
                ))
                self.list_str_tree(s, **qscpn)
        return node

    def list_str_tree_to_list(
            self, 
            tree, 
            lst,
            ):
        self.__list_str_tree_to_list(tree, lst)
        return lst

    
    def __list_str_tree_to_list(
            self, 
            node, 
            lst,
            ):
        if node.get_children():
            for child in node.get_children():
                if isinstance(child, type(node)):
                    self.utils.nested_list_insert(lst, [], node.path)
                    self.__list_str_tree_to_list(child, lst)
                if isinstance(child, list):
                    self.utils.nested_list_insert(lst, child, node.path)
        
    def str_to_list(
            self, 
            s,
            ):
        tree = self.list_str_tree(s)
        lst = []
        if self.utils.num_lists(s) == 1:
            return self.list_str_tree_to_list(tree, lst)[0]
        return self.list_str_tree_to_list(tree, lst)
    
def strcast(
        s,
        ):
    str_cast = StrCast()
    return str_cast.utils.str_cast(s)




from __future__ import annotations
from typing import Union, TypeVar, NamedTuple, Sequence, Deque, Optional, List, Type, Dict, Any
from types import ModuleType

from .strcast import strcast



CLASS = TypeVar('CLASS')

#Generic class instance type alias
INSTANCE = TypeVar('INSTANCE')

class UserInputNode(object):
    
    def __init__(self,
            xml_node: ET.Element, 
            name: str, 
            type_: str, 
            parent: Optional[UserInputNode] = None,
            ) -> None:
        self.xml_node = xml_node
        self.name = name
        self.type_ = type_ 
        self.parent = parent
        self.val = None

    def data_type(
            self, 
            xml_node,
            ):
        if not xml_node.getchildren():
            return 'non_instance'
        else:
            class_children = []
            for child in xml_node.getchildren():
                if getattr(child, 'attrib'):
                    class_children.append(child)
            if len(class_children) == len(xml_node.getchildren()):
                return 'instance_list'
            else:
                return 'instance'

    def get_name(
            self, 
            xml_node,
            ):
        return xml_node.tag 

    def add_child(
            self, 
            xml_node,
            ):
        name = self.get_name(xml_node)
        type_ = self.data_type(xml_node)
        child = UserInputNode(
            xml_node, 
            name, type_, 
            self
        )
        if type_ == 'pdt':
            child.data = strcast(xml_node.text)
        setattr(self, name, child)
        return getattr(self, name)
